/* Centaur Terminal File Server
 *
 * Version 1.0
 * 1/8/2021
 *
 * ctfs.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define PROG_HELP \
"Centaur Terminal File Server: Serve files through a terminal interface\n\
\n\
ctfs [-H] [port] &\n\
\n\
\t[-H] - Display program syntax and options\n\
\n\
\t[port] must be a valid TCP port number. Default port is 4444.\n"

#define MANIFEST_FILE	"MANIFEST"	/* File listing available files */

#define DEFAULT_PORT	4444		/* Port number if [port] is not given */

#define STRLEN		102400		/* Maximum transfer buffer (100 MB) */

char g_mesg[STRLEN];
char g_file[128];

enum
{
	_NONE,
	_FILE,
	_DIR
};

int ftype(char file[])
{
	struct stat fstat;
	int s = stat(file, &fstat);
	if (s != 0)
	{
		return _NONE;
	}
	if (S_ISREG (fstat.st_mode)) 
	{
		return _FILE;
	} else if (S_ISDIR (fstat.st_mode)) 
	{
		return _DIR;
	} else {
		return _NONE;
	}
}

int manifest(char file[])
{
	if (ftype(MANIFEST_FILE) != _FILE)
	{
		return -1;
	}

	char mline[80];
	FILE *fptr;
	fptr = fopen(MANIFEST_FILE, "r");

	while (1)
	{
		if (feof(fptr))
		{
			return 0;
		}
		fscanf(fptr, "%s", mline);
		if (mline[0] == '#')
		{
			continue;
		}
		if (!strcmp(file, mline))
		{
			return 1;
		}
	}
	fclose(fptr);
}

int vsend(int conn, char str[])
{
	send(conn, str, strlen(str), 1);
	return 0;
}

int fsend(int conn, char file[])
{
	if (ftype(file) != _FILE)
	{
		return -1;
	}
	int x;
	FILE *fptr;
	fptr = fopen(file, "r");
	for (x = 0; ; x++)
	{
		if (feof(fptr))
		{
			break;
		}
		g_mesg[x] = fgetc(fptr);
	}
	fclose(fptr);
	vsend(conn, g_mesg);
	memset(g_mesg, 0, STRLEN);
	return 0;
}

int main(int argc, char *argv[])
{
	if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 'H')
	{
		printf(PROG_HELP);
		exit(0);
	}

	if (argc > 2)
	{
		printf("Usage: %s [-H] [port] &\n", argv[0]);
		exit(1);
	}

	if (ftype(MANIFEST_FILE) != _FILE)
	{
		printf("%s: '%s' file not found in server directory\n", argv[0], MANIFEST_FILE);
		exit(1);
	}
	
	int port;
	int opt = 1;
	struct sockaddr_in addr;
	int sock;
	int addrlen = sizeof(addr);  
	int sfd = socket(AF_INET, SOCK_STREAM, 0);

	if (argc < 2)
	{
		port = DEFAULT_PORT;
	} else {
		port = atoi(argv[1]);
	}

	if (sfd == 0)
	{
		perror("socket fd");
		exit(1);
	}

	if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)))
	{
		perror("setsockopt failed");
		exit(1);
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);

	if (bind(sfd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
		perror("bind failed");
		exit(1);
	}

	if (listen(sfd, 8))
	{
		perror("listen failed");
		exit(1);
	}

	while (1)
	{
		if ((sock = accept(sfd, (struct sockaddr *)&addr, (socklen_t *)&addrlen)) < 0)
		{
			perror("accept");
			exit(1);
		}

		memset(g_file, 0, 128);

		read(sock, g_file, 128);

		if (g_file[strlen(g_file) - 1] == '\r')
		{
			g_file[strlen(g_file) - 1] = '\0';
			g_file[strlen(g_file) - 1] = '\0';
		}

		if (g_file[strlen(g_file) - 1] == '\n')
		{
			g_file[strlen(g_file) - 1] = '\0';
			g_file[strlen(g_file) - 1] = '\0';
		}

		if (!strcmp(g_file, ""))
		{
			close(sock);
		}

		if (!strcmp(g_file, "_"))
		{
			fsend(sock, MANIFEST_FILE);
		}

		if (manifest(g_file) == 1)
		{
			fsend(sock, g_file);
		}

		close(sock);

	}

	return 0;
}
